/*
* inbound.c
* incoming connection handling
*
* Copyright (c) 2004 Kathy Wang <knwang@synacklabs.net>
*
*
*/

#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <packetp.h>
#include <unistd.h>
#include "morph.h"

/* This program will accomplish the following: */
/* - Handle incoming packets */
/* - Reply to funky Nmap and similar queries */
/* - Call a packet handling function */
/* - Use the scanning tool's fingerprint files */
/* - Write to and access the state table as needed */

int 
udp_respond(struct packetp_ctx *pp_ctx,
	    struct morph_ctx *morph_handle, uint8_t *in_packet)
{
  uint8_t out_packet[IP_LEN_MAX];
  struct ip_hdr *ip_header_in;
  struct udp_hdr *udp_header_in;
  struct ip_hdr *ip_header_out;
  struct icmp_hdr *icmp_header_out;
  struct icmp_msg_quote *out_unreach;
  int in_pkt_sz;
  
  memset(out_packet, 0, IP_LEN_MAX);
  ip_header_out=(struct ip_hdr *)out_packet;
  icmp_header_out=(struct icmp_hdr *)(out_packet+20);
  ip_header_in=(struct ip_hdr *)in_packet;
  udp_header_in=(struct udp_hdr *)(in_packet+(4*ip_header_in->ip_hl));
  out_unreach = (struct icmp_msg_quote *)
    (((uint8_t *)icmp_header_out) + ICMP_HDR_LEN);

  ip_header_out->ip_v=4;
  ip_header_out->ip_hl=5;
  ip_header_out->ip_p=IP_PROTO_ICMP;
  if(pp_ctx->simplified && (pp_ctx->wedge_type == PP_FAKEIP)) {
    memcpy(&(ip_header_out->ip_src), &(pp_ctx->proxy_ip.addr_ip), 4);
    memcpy(&(ip_header_out->ip_dst), &(pp_ctx->target_ip.addr_ip), 4);
  } else {
    memcpy(&(ip_header_out->ip_src), &(ip_header_in->ip_dst), 4);
    memcpy(&(ip_header_out->ip_dst), &(ip_header_in->ip_src), 4);
  }
  icmp_header_out->icmp_type = ICMP_UNREACH;
  icmp_header_out->icmp_code = ICMP_UNREACH_PORT;

  switch(morph_handle->os_type) {
  case OS_TYPE_WIN2K:
    ip_header_out->ip_id=htons(morph_handle->ip_id++);
    ip_header_out->ip_off=0;
    ip_header_out->ip_ttl=128;
    /*    ip_header_in->ip_len = htons(0x148); */
    memcpy(&(out_unreach->icmp_ip), in_packet, 12+(4*ip_header_in->ip_hl));
    ip_header_out->ip_len = htons(12+(4*ip_header_in->ip_hl) +
				  IP_HDR_LEN +
				  ICMP_HDR_LEN);    
    break;

  case OS_TYPE_OPENBSD_33:
    ip_header_out->ip_id=rand_uint16(morph_handle->randomizer);
    ip_header_out->ip_ttl=255;
    ip_header_in->ip_len = htons(0x134); 
    memcpy(&(out_unreach->icmp_ip), in_packet, 12+(4*ip_header_in->ip_hl));
    ip_header_out->ip_len = htons(12+(4*ip_header_in->ip_hl) +
				  IP_HDR_LEN +
				  ICMP_HDR_LEN);
    break;

  case OS_TYPE_LINUX_24:
    ip_header_out->ip_id=rand_uint16(morph_handle->randomizer);
    ip_header_out->ip_ttl=64;
    ip_header_out->ip_tos=IP_TOS_PREC_INTERNETCONTROL;
    in_pkt_sz = ntohs(ip_header_in->ip_len);
    if(in_pkt_sz > 72)
      in_pkt_sz = 72;
    memcpy(&(out_unreach->icmp_ip), in_packet, 
	   in_pkt_sz + (4*ip_header_in->ip_hl));
    ip_header_out->ip_len = htons(in_pkt_sz +
				  (4*ip_header_in->ip_hl) +
				  IP_HDR_LEN +
				  ICMP_HDR_LEN);
    break;
  }
  ip_checksum(out_packet,ntohs(ip_header_out->ip_len));
  packetp_packet_inject(pp_ctx, out_packet);
  return(0);
}

int 
icmp_respond(struct packetp_ctx *pp_ctx, struct morph_ctx *morph_handle, 
	     uint8_t *in_packet)
{
  uint8_t out_packet[IP_LEN_MAX];
  uint8_t *byte_ptr;
  int i;
  int j;
  int sentbytes;
  intf_t *intf_handle;
  struct ip_hdr *ip_header_in;
  struct icmp_hdr *icmp_header_in;
  struct ip_hdr *ip_header_out;
  struct icmp_hdr *icmp_header_out;
  struct addr tmp_address;
  struct intf_entry interface;
  struct icmp_msg_echo *in_echo;
  struct icmp_msg_echo *out_echo;
  struct icmp_msg_tstamp *out_tstamp;
  struct icmp_msg_tstamp *in_tstamp;
  struct icmp_msg_mask *in_mask;
  struct icmp_msg_mask *out_mask;
  struct timeval now;

  memset(out_packet, 0, IP_LEN_MAX);
  ip_header_out=(struct ip_hdr *)out_packet;
  icmp_header_out=(struct icmp_hdr *)(out_packet+20);
  ip_header_in=(struct ip_hdr *)in_packet;
  icmp_header_in=(struct icmp_hdr *)(in_packet+(4*ip_header_in->ip_hl));

  ip_header_out->ip_v=4;
  ip_header_out->ip_hl=5;
  ip_header_out->ip_p=IP_PROTO_ICMP;
  memcpy(&(ip_header_out->ip_src), &(ip_header_in->ip_dst), 4);
  memcpy(&(ip_header_out->ip_dst), &(ip_header_in->ip_src), 4);

  printf("pkts: %08x %08x %d\n", out_packet, in_packet, ip_header_in->ip_len);

  switch(icmp_header_in->icmp_type) {
  case ICMP_ECHO:
    in_echo=(struct icmp_msg_echo *)(((uint8_t *)icmp_header_in)+ICMP_HDR_LEN);
    out_echo=(struct icmp_msg_echo *)
      (((uint8_t *)icmp_header_out)+ICMP_HDR_LEN);
    icmp_header_out->icmp_type = ICMP_ECHOREPLY;
    icmp_header_out->icmp_code = ICMP_CODE_NONE;
    out_echo->icmp_id=in_echo->icmp_id;
    out_echo->icmp_seq=in_echo->icmp_seq;
    memcpy(&(out_echo->icmp_data), &(in_echo->icmp_data), (ntohs(ip_header_in->ip_len))-(ip_header_in->ip_hl*4)-(ICMP_HDR_LEN)-4);
    ip_header_out->ip_len=ip_header_in->ip_len;
    break;

  case ICMP_TSTAMP:
    in_tstamp=(struct icmp_msg_tstamp *)
      (((uint8_t *)icmp_header_in)+ICMP_HDR_LEN);
    out_tstamp=(struct icmp_msg_tstamp *)
      (((uint8_t *)icmp_header_out)+ICMP_HDR_LEN);
    icmp_header_out->icmp_type = ICMP_TSTAMPREPLY;
    icmp_header_out->icmp_code = ICMP_CODE_NONE;    
    out_tstamp->icmp_id=in_tstamp->icmp_id;
    out_tstamp->icmp_seq=in_tstamp->icmp_seq;
    gettimeofday(&now, NULL);
    out_tstamp->icmp_ts_orig = in_tstamp->icmp_ts_orig;
    out_tstamp->icmp_ts_rx = htonl(now.tv_sec);
    out_tstamp->icmp_ts_tx = htonl(now.tv_sec);
    ip_header_out->ip_len=ip_header_in->ip_len;
    break;

  case ICMP_MASK:
    in_mask=(struct icmp_msg_mask *)(((uint8_t *)icmp_header_in)+ICMP_HDR_LEN);
    out_mask=(struct icmp_msg_mask *)
      (((uint8_t *)icmp_header_out)+ICMP_HDR_LEN);
    icmp_header_out->icmp_type = ICMP_MASKREPLY;
    icmp_header_out->icmp_code = ICMP_CODE_NONE;
    out_mask->icmp_id=in_mask->icmp_id;
    out_mask->icmp_seq=in_mask->icmp_seq;
    intf_handle = intf_open();
    memcpy(&(ip_header_in->ip_src), &(tmp_address.addr_ip), IP_ADDR_LEN);
    tmp_address.addr_type = ADDR_TYPE_IP;
    tmp_address.addr_bits = IP_ADDR_BITS;
    interface.intf_len = sizeof(interface);
    intf_get_dst(intf_handle, &interface, &tmp_address);
    byte_ptr = (uint8_t *)(&(out_mask->icmp_mask));
    j = interface.intf_addr.addr_bits/8;
    for(i = 0; i < j; i++) {
      *byte_ptr = 0xFF;
      byte_ptr++;
    }
    i = interface.intf_addr.addr_bits%8;
    *byte_ptr = ~0 << (8 - i);
    ip_header_out->ip_len = htons(IP_HDR_LEN+
				  ICMP_HDR_LEN+
				  sizeof(struct icmp_msg_tstamp));
    break;
    
  case ICMP_INFO:
    /* 
     * Umm, this is marked "Obsolete" by Stevens, and none of our supported
     * OSes respond to it. We can add it in if we support AIX in the future,
     * though. ;-)
     */
    break;
  }

  
  switch(morph_handle->os_type) {
  case OS_TYPE_WIN2K:
    ip_header_out->ip_id=htons(morph_handle->ip_id++);
    ip_header_out->ip_off=htons(IP_DF);
    ip_header_out->ip_ttl=128;
    switch(icmp_header_in->icmp_type) {
    case ICMP_MASK:
    case ICMP_INFO:
      return(1);
    }      
    break;

  case OS_TYPE_OPENBSD_33:
    ip_header_out->ip_id=rand_uint16(morph_handle->randomizer);
    ip_header_out->ip_off=htons(IP_DF);
    ip_header_out->ip_ttl=255;
    switch(icmp_header_in->icmp_type) {
    case ICMP_MASK:
    case ICMP_INFO:
      return(1);
    }
    break;

  case OS_TYPE_LINUX_24:
    ip_header_out->ip_id=0;
    ip_header_out->ip_off=htons(IP_DF);
    ip_header_out->ip_ttl=64;
    switch(icmp_header_in->icmp_type) {
    case ICMP_MASK:
    case ICMP_INFO:
      return(1);
    }
    break;

  }
  ip_checksum(out_packet,ntohs(ip_header_out->ip_len));
  sentbytes = packetp_packet_inject(pp_ctx, out_packet);

  return(0);
}

int 
inbound(struct packetp_ctx *pp_ctx, uint8_t *packet, void *void_ctx)
{
  struct morph_ctx *morph_handle;
  struct morph_session *session;
  struct ip_hdr *ip_header;
  struct tcp_hdr *tcp_header;
  struct icmp_hdr *icmp_header;
  struct udp_hdr *udp_header;
  uint32_t ack_num;
  uint8_t *tmp_loc;
  struct tcp_opt *option;
  int i;

  morph_handle=(struct morph_ctx *)void_ctx;

  ip_header=(struct ip_hdr *)packet;
  if (ip_header->ip_p==IP_PROTO_TCP) {
    
    tcp_header=(struct tcp_hdr *)(packet+(ip_header->ip_hl*4));

    printf ("Inbound: src %08x dst %08x sport %d dport %d flags %02x\n", ntohl(ip_header->ip_src), ntohl(ip_header->ip_dst), ntohs(tcp_header->th_sport), ntohs(tcp_header->th_dport), tcp_header->th_flags);

    if(morph_port_open(IP_PROTO_TCP, ntohs(tcp_header->th_dport))==0) {
      tcp_closed_respond(pp_ctx, morph_handle, packet);
      return(1);
    }
    
    session = morph_session_find(packet, morph_handle, IN, pp_ctx);

    if(session == NULL) {
      if(tcp_header->th_flags & TH_SYN) {
	session = morph_session_add(packet, morph_handle, IN, pp_ctx);
	if(session != NULL)
	  session->tcp_state = TCP_SYN_RECEIVED;

	printf("New inbound session, sport %d dport %d flags %08x\n", ntohs(tcp_header->th_sport), ntohs(tcp_header->th_dport),tcp_header->th_flags);
	return(0);
      } else if(tcp_header->th_flags & TH_RST) {
	return(0);
      } else {
	tcp_open_respond(pp_ctx, morph_handle, packet);
	return(1);
      }
    } else {
      ack_num = ntohl(tcp_header->th_ack) + session->seq_offset;
      printf("in %d %d %d\n", ntohl(tcp_header->th_ack), session->seq_offset, ack_num);
      tcp_header->th_ack = htonl(ack_num);
    }

    if (tcp_header->th_flags & TH_RST) {
      morph_session_del(morph_handle, pp_ctx);
    } else if (tcp_header->th_flags == (TH_SYN|TH_ACK)) {
      if(session->tcp_state == TCP_SYN_SENT) {
	session->tcp_state = TCP_SYNACK_RECEIVED;
      } else {
	tcp_open_respond(pp_ctx, morph_handle, packet);
	return(1);
      }
    } else if (tcp_header->th_flags & TH_ACK) {
      if (session->tcp_state == TCP_SYNACK_SENT) {
	session->tcp_state = TCP_ESTABLISHED;
      }
      if(tcp_header->th_flags & TH_FIN) {
	if (session->tcp_state == TCP_FINACK_SENT){
	  morph_session_del(morph_handle, pp_ctx);
	} else {
	  session->tcp_state = TCP_FINACK_RECEIVED;
	}
      } else {
	if (session->tcp_state == TCP_FINACK_SENT) {
	  morph_session_del(morph_handle, pp_ctx);
	}
      }
    } else if (tcp_header->th_flags & TH_FIN) {
      if(session->tcp_state == TCP_ESTABLISHED) {
	session->tcp_state = TCP_FIN_RECEIVED;
      }
    }
  } else if (ip_header->ip_p==IP_PROTO_ICMP) {

    icmp_header=(struct icmp_hdr *)(packet+(ip_header->ip_hl*4));

    switch(icmp_header->icmp_type) {
    case ICMP_ECHO:
    case ICMP_MASK:
    case ICMP_INFO:
    case ICMP_TSTAMP:
      icmp_respond(pp_ctx, morph_handle, packet);
      return(1);
      break;
      
    case ICMP_UNREACH:
      return(0);
      break;
      
    }
  } else if (ip_header->ip_p==IP_PROTO_UDP) {
    udp_header=(struct udp_hdr *)(packet+(ip_header->ip_hl*4));
    if(morph_port_open(IP_PROTO_UDP, ntohs(udp_header->uh_dport))==0) {
      udp_respond(pp_ctx, morph_handle, packet);
      return(1);
    }
  }
  return(0);
}




