/*
 * main.c
 *
 * Includes main logic loop
 *
 * Copyright (c) 2004 Kathy Wang <knwang@synacklabs.net>
 */

#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <packetp.h>
#include <unistd.h>
#include "morph.h"

/* usage: prints the usage string and exits */

void
usage(char *prog_name)
{
  printf("Usage: %s -o <os_type> [-p <proxyIP>][-t <target address>]\n", prog_name);
  exit(0);
}

int
main (int argc, char *argv[])
{
  struct packetp_ctx *pp_ctx;
  struct morph_ctx m_ctx;
  int counter;
  int got_proxy;
  int got_target;

  got_proxy = 0;
  got_target = 0;

  if(getuid() != 0) {
    fprintf(stderr, "Gotta be root to run %s\n", argv[0]);
    return(-1);
  }


  if (argc < 2) {
    usage (argv[0]);
  }
  
  m_ctx.randomizer=rand_open();
  m_ctx.ip_handle=ip_open();
  m_ctx.sessions = hcreate(8);

  if((pp_ctx = packetp_init()) == NULL) 
    return(-1);

  while((counter=getopt(argc, argv, "o:p:t:")) != -1) {
    switch(counter) {
    case 'p':
      printf("proxy %s\n", optarg);
      if ((addr_pton(optarg, &(pp_ctx->proxy_ip))) < 0) {
        fprintf(stderr, "Cannot resolve proxy IP: %s\n", optarg);
        return(-1);
      }
      got_proxy = 1;
      break;

    case 't':
      printf("target %s\n", optarg);
      if ((addr_pton(optarg, &(pp_ctx->target_ip))) < 0) {
        fprintf(stderr, "Cannot resolve target IP: %s\n", optarg);
        return(-1);
      }
      got_target = 1;
      break;

    case 'o':
      printf("os %s\n", optarg);
      if (strstr(optarg,"win2k")) {
	m_ctx.os_type=OS_TYPE_WIN2K;
      } else if (strstr(optarg, "openbsd")) {
	m_ctx.os_type=OS_TYPE_OPENBSD_33;
      } else if (strstr(optarg, "linux")) {
	m_ctx.os_type=OS_TYPE_LINUX_24;
      }
      break;
    }
  } 

  if(got_proxy) {
    if(!(got_target)) {
      fprintf(stderr, 
              "Must specify remote IP address when specifying a proxy IP\n");
      usage(argv[0]);
    }
    if(pp_ctx->target_ip.addr_bits < IP_ADDR_BITS) {
      fprintf(stderr, 
              "Remote IP address cannot be CIDR block if proxy mode used\n"); 
      usage(argv[0]);
    }
    pp_ctx->wedge_type = PP_FAKEIP;
  }

  packetp_start(pp_ctx, inbound, outbound, &m_ctx);
  return(0);
}



