/*
 * morph.h
 *
 * Structures and definitions for Morph.
 *
 * Portions stolen from stegtunnel.h, Copyright (c) Todd MacDermid <tmacd@synacklabs.net>
 *
 * Additional code Copyright (c) 2004 Kathy Wang <knwang@synacklabs.net>
 *
 */

#ifndef MORPH_H
#define MORPH_H

#include "hashtab.h"

/* 
 * tcp_state definitions. Note that these don't correspond identically
 * to the TCP FSM states in dnet/tcp.h, because we're really proxying
 * for another state machine. We have no concept, for example, of
 * TCP_LISTEN, we just forward on. Also, we can't assume we've ACKed
 * anything we've received, that takes place in two steps. Fun, ne?
 */

#define TCP_NO_STATE          0
#define TCP_SYN_SENT          1
#define TCP_SYN_RECEIVED      2
#define TCP_SYNACK_SENT       3
#define TCP_SYNACK_RECEIVED   4
#define TCP_ESTABLISHED       5
#define TCP_FIN_SENT          6
#define TCP_FIN_RECEIVED      7
#define TCP_FINACK_SENT       8
#define TCP_FINACK_RECEIVED   9
#define TCP_SENT_CLOSED       10
#define TCP_RECV_CLOSED       11


int morph_port_open(uint8_t protocol, uint16_t port);

struct morph_ctx {
  uint32_t os_type;
  uint16_t ip_id;
  uint32_t last_seq;
  rand_t *randomizer;
  ip_t *ip_handle;
  htab *sessions;
};

#define OS_TYPE_WIN2K      1
#define OS_TYPE_OPENBSD_33 2
#define OS_TYPE_LINUX_24   3


/*
 * morph_session objects are allocated one per TCP connection, and are
 * deallocated when that session ends (either through FIN or RST).
 */

struct morph_session {
  struct addr remote_ip;
  int32_t seq_offset;
  uint32_t timestamp;
  int tcp_state;          /* Defined above */
  uint16_t proxy_port;    /* Used in PP_FAKEIP mode */
  uint16_t remote_port;
  uint16_t local_port;
};

struct morph_session *morph_session_add(uint8_t *, struct morph_ctx *,
					int, struct packetp_ctx *);

void morph_session_del(struct morph_ctx *, struct packetp_ctx *);


struct morph_session *morph_session_find(uint8_t *, struct morph_ctx *, 
					 int, struct packetp_ctx *);

int tcp_open_respond(struct packetp_ctx *, struct morph_ctx *, uint8_t *);
int tcp_closed_respond(struct packetp_ctx *, struct morph_ctx *, uint8_t *);

int outbound(struct packetp_ctx *, uint8_t *, void *);
int inbound(struct packetp_ctx *, uint8_t *, void *);
#endif /* MORPH_H */
