#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <packetp.h>
#include <unistd.h>
#include "morph.h"

void
tcp_option_reorder(struct morph_ctx *morph_handle, uint8_t *packet)
{
  uint8_t *tmp_dst;
  uint8_t *tmp_loc;
  struct ip_hdr *ip_header;
  struct tcp_hdr *tcp_header;
  struct tcp_opt *option;
  struct tcp_opt timestamp;
  int i;
  int have_ts;

  have_ts = 0;

  ip_header=(struct ip_hdr *)packet;
  tcp_header=(struct tcp_hdr *)(packet+(ip_header->ip_hl*4));

  tmp_loc = (uint8_t *) (((uint8_t *)tcp_header) + 20);
  i = 20;
  while(i < (tcp_header->th_off * 4)) {
    option = (struct tcp_opt *)(((uint8_t *)tcp_header) + i);
    if(option->opt_type == TCP_OPT_EOL) 
      break;
    if(option->opt_type == TCP_OPT_NOP) {
      i++;
    } else {
      if(option->opt_type == TCP_OPT_TIMESTAMP) {
	memcpy(&timestamp, option, option->opt_len);
	have_ts = 1;
      }
      i += option->opt_len;
    }
  }
  
  switch(morph_handle->os_type) {
  case OS_TYPE_WIN2K:
    tmp_dst = (uint8_t *) (((uint8_t *)tcp_header) + 40);
    memmove(tmp_dst, tmp_loc, 
	    (ntohs(ip_header->ip_len) - 
	     (tcp_header->th_off*4) - 
	     (ip_header->ip_hl *4)));
    tmp_loc = (uint8_t *) (((uint8_t *)tcp_header) + 20);
    ip_header->ip_len = htons(ntohs(ip_header->ip_len) + ((10 - tcp_header->th_off) *4));
    tcp_header->th_off = 10;
    option = (struct tcp_opt *)tmp_loc;
    option->opt_type = TCP_OPT_MSS;
    option->opt_len = 4;
    option->opt_data.mss = htons(1460);
    
    tmp_loc += 4;
    *tmp_loc = TCP_OPT_NOP;
    
    tmp_loc++;
    option = (struct tcp_opt *)tmp_loc;
    option->opt_type = TCP_OPT_WSCALE;
    option->opt_len = 3;
    option->opt_data.wscale = 0;
    
    tmp_loc += 3;
    *tmp_loc = TCP_OPT_NOP;
    
    tmp_loc++;
    *tmp_loc = TCP_OPT_NOP;
    
    tmp_loc++;
    if(have_ts) {
      memcpy(tmp_loc, &timestamp, 10);
    } else {
      option = (struct tcp_opt *)tmp_loc;
      option->opt_type = TCP_OPT_TIMESTAMP;
      option->opt_len = 10;
    }
    break;
      
  case OS_TYPE_OPENBSD_33:
    
    tmp_dst = (uint8_t *) (((uint8_t *)tcp_header) + 40);
	
    memmove(tmp_dst, tmp_loc, 
	    (ntohs(ip_header->ip_len) - 
	     (tcp_header->th_off*4) - 
	     (ip_header->ip_hl *4))); 
    
    ip_header->ip_len = htons(ntohs(ip_header->ip_len) + ((10 - tcp_header->th_off) *4)); 
    tcp_header->th_off = 10;
    
    tmp_loc = (uint8_t *) (((uint8_t *)tcp_header) + 20); 

    option = (struct tcp_opt *)tmp_loc;
    option->opt_type = TCP_OPT_MSS;
    option->opt_len = 4;
    option->opt_data.mss = htons(1460);
    
    tmp_loc += 4;
    
    *tmp_loc = TCP_OPT_NOP;
	
    tmp_loc++;
    option = (struct tcp_opt *)tmp_loc;
    option->opt_type = TCP_OPT_WSCALE;
    option->opt_len = 3;
    option->opt_data.wscale = 0;
    
    tmp_loc += 3;
    *tmp_loc = TCP_OPT_NOP;
    
    tmp_loc++;
    *tmp_loc = TCP_OPT_NOP;
    
    tmp_loc++;
    if(have_ts) {
      memcpy(tmp_loc, &timestamp, 10);
    } else {
      option = (struct tcp_opt *)tmp_loc;
      option->opt_type = TCP_OPT_TIMESTAMP;
      option->opt_len = 10;
    }
    break;
    
  }
  return;
}

int 
outbound(struct packetp_ctx *pp_ctx, uint8_t *packet, void *void_ctx)
{
  struct morph_ctx *morph_handle;
  struct morph_session *session;
  struct ip_hdr *ip_header;
  struct tcp_hdr *tcp_header;
  uint8_t *tmp_loc;
  uint8_t *tmp_dst;
  uint32_t seq_num;

  morph_handle=(struct morph_ctx *)void_ctx;

  ip_header=(struct ip_hdr *)packet;
  if (ip_header->ip_p==IP_PROTO_TCP) {
    tcp_header=(struct tcp_hdr *)(packet+(ip_header->ip_hl*4));
    printf ("Outbound: src %08x dst %08x sport %d dport %d flags %02x\n", ntohl(ip_header->ip_src), ntohl(ip_header->ip_dst), ntohs(tcp_header->th_sport), ntohs(tcp_header->th_dport), tcp_header->th_flags);
    
    session = morph_session_find(packet, morph_handle, OUT, pp_ctx);

    if(session == NULL) {
      if(tcp_header->th_flags == TH_SYN) {
	session = morph_session_add(packet, morph_handle, OUT, pp_ctx);
	if(session != NULL)
	  session->tcp_state = TCP_SYN_RECEIVED;
      } else {
	fprintf(stderr, "Non-SYN unknown session: src %08x dst %08x sport %d dport %d flags %02x\n", ntohl(ip_header->ip_src), ntohl(ip_header->ip_dst), ntohs(tcp_header->th_sport), ntohs(tcp_header->th_dport), tcp_header->th_flags);
      }
    } else {
      
      if ((tcp_header->th_flags == (TH_SYN|TH_ACK)) && 
          (session->tcp_state == TCP_SYN_RECEIVED)) {
        session->tcp_state = TCP_SYNACK_SENT;
	seq_num = new_sequence(morph_handle); 
	/*	seq_num = ntohl(tcp_header->th_seq); */

        session->seq_offset = ntohl(tcp_header->th_seq) - 
          seq_num;
	printf("gn %d %d %d\n", ntohl(tcp_header->th_seq), session->seq_offset, seq_num);
        tcp_header->th_seq = htonl(seq_num); 
	
	
      } else {
	seq_num = ntohl(tcp_header->th_seq) - session->seq_offset;
	      printf("ut %d %d %d\n", ntohl(tcp_header->th_seq), session->seq_offset, seq_num);
	tcp_header->th_seq = htonl(seq_num);
      }
      

      if (tcp_header->th_flags & TH_RST) {
	/*	morph_session_del(morph_handle, pp_ctx); */
      } else if (tcp_header->th_flags == (TH_SYN|TH_ACK)) {

      }  else if (tcp_header->th_flags & TH_ACK) {
	if(tcp_header->th_flags & TH_SYN) {
	  if(session->tcp_state == TCP_SYN_RECEIVED) {
	    session->tcp_state = TCP_SYNACK_RECEIVED;
	  } else if (session->tcp_state == TCP_SYNACK_RECEIVED) {
	    session->tcp_state = TCP_ESTABLISHED;
	  }
	} else if(tcp_header->th_flags & TH_FIN) {
	  if (session->tcp_state == TCP_FINACK_RECEIVED){
	    morph_session_del(morph_handle, pp_ctx);
	  } else {
	    session->tcp_state = TCP_FINACK_SENT;
	  }
	} else {
	  if(session->tcp_state == TCP_SYNACK_RECEIVED) {
	    session->tcp_state = TCP_ESTABLISHED;
	  } else if (session->tcp_state == TCP_FINACK_RECEIVED) {
	    morph_session_del(morph_handle, pp_ctx);
	  } 
	}
      } else if (tcp_header->th_flags & TH_FIN) {
	if(session->tcp_state == TCP_ESTABLISHED) {
	  session->tcp_state = TCP_FIN_SENT;
	}
      }
    }

    tmp_loc = (uint8_t *)(((uint8_t *)tcp_header)+(tcp_header->th_off*4));

    switch(morph_handle->os_type) {
    case OS_TYPE_WIN2K:
      ip_header->ip_id=htons(morph_handle->ip_id++);
      ip_header->ip_off=htons(IP_DF);
       ip_header->ip_ttl=128;
       tcp_header->th_win = htons(0x402E);

       
      if(tcp_header->th_flags == (TH_SYN|TH_ACK)) {
	tcp_option_reorder(morph_handle, packet);
      }

      break;
      
    case OS_TYPE_OPENBSD_33:
      ip_header->ip_id=rand_uint16(morph_handle->randomizer);
      ip_header->ip_ttl=255;
      tcp_header->th_win = htons(0x403D);
      ip_header->ip_off=htons(IP_DF);
      
      
      if(tcp_header->th_flags == (TH_SYN|TH_ACK)) {
	tcp_option_reorder(morph_handle, packet);
      }     
      break;
      
    case OS_TYPE_LINUX_24:
      ip_header->ip_id=0;
      ip_header->ip_ttl=64;
      ip_header->ip_off=htons(IP_DF);
      tcp_header->th_win = htons(5792);
      break;
    }
    
  }
  
  return(0);
}

