#include <dnet.h>
#include <stdio.h>

#define LINUX_TCP "/proc/net/tcp"
#define LINUX_UDP "/proc/net/udp"
#define MAX_LINE 1024
#define STATE_LISTEN 0x0A
#define UDP_OPEN 0x07
#define LINUX 1

int
morph_port_open(uint8_t protocol, uint16_t port)
{
  FILE *f_handle;
  char *separator_loc;
  char in_line[MAX_LINE];
  uint32_t tmp_addr;
  uint32_t tmp_port;
  uint32_t tmp_state;

#ifdef LINUX
  if(protocol == IP_PROTO_TCP) {
    if((f_handle = fopen(LINUX_TCP, "r")) == NULL) {
      fprintf(stderr, 
	      "morph_port_open: Could not open tcp state file: %s\n", 
	      LINUX_TCP);
      return(-1);
    }
    
  } else if (protocol == IP_PROTO_UDP) {
    if((f_handle = fopen(LINUX_UDP, "r")) == NULL) {
      fprintf(stderr, 
	      "morph_port_open: Could not open udp state file: %s\n", 
	      LINUX_UDP);
      return(-1);
    }
  } else {
    fprintf(stderr, "morph_port_open: Unknown protocol\n");
  }

  fgets(in_line, MAX_LINE-1, f_handle); /* Throw away first line */
  while(fgets(in_line, MAX_LINE-1, f_handle) != NULL) {
    if((separator_loc = strchr(in_line, ':')) != NULL){
      if((sscanf(separator_loc, 
		 ": %x:%x %*x:%*x %x", 
		 &tmp_addr,
		 &tmp_port,
		 &tmp_state)) == 0) {
	fprintf(stderr, "morph_port_open: sscanf failure\n");
      } else {
	if(((tmp_state == STATE_LISTEN) || (tmp_state == UDP_OPEN))
	  && tmp_port == port) {
	  return(1);
	}
      }
    }
  }
  fclose(f_handle);
  return(0);
#endif
}

