#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <packetp.h>
#include <unistd.h>
#include <netinet/in.h>
#include "morph.h"

uint32_t 
new_sequence(struct morph_ctx *morph_handle)
{
  uint32_t new_sequence;
  
  
  switch(morph_handle->os_type) {
  case OS_TYPE_WIN2K:
    new_sequence = rand_uint32(morph_handle->randomizer);
    new_sequence = new_sequence%0x1E438;
    new_sequence += 0x1517;
    new_sequence += morph_handle->last_seq;
    break;

  case OS_TYPE_OPENBSD_33:
    new_sequence = rand_uint32(morph_handle->randomizer);

    break;

  case OS_TYPE_LINUX_24:
    /* Look at nmap fingerprint for Linux seq number */
    /* OK, Nmap is doing TCP timestamp sequence predictions */
    /* that's challenging, let's just focus on Windows for now */

    break;
  }

  morph_handle->last_seq = new_sequence;
  printf("%08x\n", new_sequence);
  return(new_sequence);
}

int 
tcp_open_respond(struct packetp_ctx *pp_ctx, 
		 struct morph_ctx *morph_handle, uint8_t *in_packet)
{
  uint8_t out_packet[IP_LEN_MAX];
  struct ip_hdr *ip_header_in;
  struct tcp_hdr *tcp_header_in;
  struct ip_hdr *ip_header_out;
  struct tcp_hdr *tcp_header_out;
  int sentbytes;
  int respond = 0;
  uint8_t *tmp_loc;
  uint8_t *tmp_dst;
  struct tcp_opt *option;
  int i;

  memset(out_packet, 0, IP_LEN_MAX);
  ip_header_out=(struct ip_hdr *)out_packet;
  tcp_header_out=(struct tcp_hdr *)(out_packet+20);
  ip_header_in=(struct ip_hdr *)in_packet;
  tcp_header_in=(struct tcp_hdr *)
    (((uint8_t *)in_packet)+(4*ip_header_in->ip_hl));

  ip_header_out->ip_v=4;
  ip_header_out->ip_hl=5;
  ip_header_out->ip_p=IP_PROTO_TCP;
  ip_header_out->ip_len=htons(IP_HDR_LEN+TCP_HDR_LEN);

  if(pp_ctx->simplified && (pp_ctx->wedge_type == PP_FAKEIP)) {
    memcpy(&(ip_header_out->ip_src), &(pp_ctx->proxy_ip.addr_ip), 4);
    memcpy(&(ip_header_out->ip_dst), &(pp_ctx->target_ip.addr_ip), 4);
  } else {
    memcpy(&(ip_header_out->ip_src), &(ip_header_in->ip_dst), 4);
    memcpy(&(ip_header_out->ip_dst), &(ip_header_in->ip_src), 4);
  }

  tcp_header_out->th_sport = tcp_header_in->th_dport;
  tcp_header_out->th_dport = tcp_header_in->th_sport;

  tcp_header_out->th_flags = TH_RST;

  printf("open respond port %d flags %02x\n", ntohs(tcp_header_in->th_dport),
	 tcp_header_in->th_flags);

  switch(morph_handle->os_type) {
  case OS_TYPE_WIN2K:
    ip_header_out->ip_id=htons(morph_handle->ip_id++);
    ip_header_out->ip_off=htons(IP_DF);
    ip_header_out->ip_off=0;
    ip_header_out->ip_ttl=128;
    tcp_header_out->th_win = htons(65535);
    if (tcp_header_in->th_flags == 0) {
      tcp_header_out->th_seq = htonl(new_sequence(morph_handle));
      tcp_header_out->th_win = 0;
      tcp_header_out->th_ack = tcp_header_in->th_seq;
      ip_header_out->ip_off = 0;
      tcp_header_out->th_flags = TH_RST|TH_ACK;
      respond=1;
    }
    if (tcp_header_in->th_flags & TH_ACK) {
      tcp_header_out->th_win = 0;
      ip_header_out->ip_off = 0;
      tcp_header_out->th_flags = TH_RST;
      respond=1;
    }
    
    break;
  case OS_TYPE_OPENBSD_33:
    ip_header_out->ip_id=rand_uint16(morph_handle->randomizer);
    ip_header_out->ip_ttl=255;
    tcp_header_out->th_win = htons(0x403D);
    ip_header_out->ip_off=htons(IP_DF);
    if(tcp_header_in->th_flags & TH_ACK) {
      tcp_header_out->th_win = htons(0x4000);
      respond = 1;
    }
    if(tcp_header_in->th_flags & TH_SYN) {
      tcp_header_out->th_ack = htonl(ntohl(tcp_header_in->th_seq) + 1); 
      tcp_header_out->th_flags = TH_SYN|TH_ACK;
      respond = 1;
      tcp_header_out->th_off = 10; 
      tmp_loc = (uint8_t *) (((uint8_t *)tcp_header_out) + 20);
      option = (struct tcp_opt *)tmp_loc;
      option->opt_type = TCP_OPT_MSS;
      option->opt_len = 4;
      option->opt_data.mss = htons(1460);
      tmp_loc += 4;
      *tmp_loc = TCP_OPT_NOP;
      tmp_loc++;
      option = (struct tcp_opt *)tmp_loc;
      option->opt_type = TCP_OPT_WSCALE;
      option->opt_len = 3;
      option->opt_data.wscale = 0;
      tmp_loc += 3;
      *tmp_loc = TCP_OPT_NOP;
      tmp_loc++;
      *tmp_loc = TCP_OPT_NOP;
      tmp_loc++;
      option = (struct tcp_opt *)tmp_loc;
      option->opt_type = TCP_OPT_TIMESTAMP;
      option->opt_len = 10;
    }
    break;

  case OS_TYPE_LINUX_24:
    ip_header_out->ip_id=rand_uint16(morph_handle->randomizer);
    ip_header_out->ip_off=htons(IP_DF);
    ip_header_out->ip_ttl=64;
    tcp_header_out->th_win = htons(5792);
    break;
  }

  if(respond) {
    ip_checksum(out_packet,ntohs(ip_header_out->ip_len));
    sentbytes = packetp_packet_inject(pp_ctx, out_packet);
  }

  return(0);
}

int 
tcp_closed_respond(struct packetp_ctx *pp_ctx, 
		   struct morph_ctx *morph_handle, uint8_t *in_packet)
{
  uint8_t out_packet[IP_LEN_MAX];
  struct ip_hdr *ip_header_in;
  struct tcp_hdr *tcp_header_in;
  struct ip_hdr *ip_header_out;
  struct tcp_hdr *tcp_header_out;
  int sentbytes;
  
  memset(out_packet, 0, IP_LEN_MAX);
  ip_header_out=(struct ip_hdr *)out_packet;
  tcp_header_out=(struct tcp_hdr *)(out_packet+20);
  ip_header_in=(struct ip_hdr *)in_packet;
  tcp_header_in=(struct tcp_hdr *)
    (((uint8_t *)in_packet)+(4*ip_header_in->ip_hl));

  ip_header_out->ip_v=4;
  ip_header_out->ip_hl=5;
  ip_header_out->ip_p=IP_PROTO_TCP;
  ip_header_out->ip_len=htons(IP_HDR_LEN+TCP_HDR_LEN);
  if(pp_ctx->simplified && (pp_ctx->wedge_type == PP_FAKEIP)) {
    memcpy(&(ip_header_out->ip_src), &(pp_ctx->proxy_ip.addr_ip), 4);
    memcpy(&(ip_header_out->ip_dst), &(pp_ctx->target_ip.addr_ip), 4);
  } else {
    memcpy(&(ip_header_out->ip_src), &(ip_header_in->ip_dst), 4);
    memcpy(&(ip_header_out->ip_dst), &(ip_header_in->ip_src), 4);
  }

  tcp_header_out->th_sport = tcp_header_in->th_dport;
  tcp_header_out->th_dport = tcp_header_in->th_sport;
  
  printf("closed respond port %d flags %02x\n", ntohs(tcp_header_in->th_dport),
	 tcp_header_in->th_flags);

  
  tcp_header_out->th_flags = TH_RST;

  switch(morph_handle->os_type) {
  case OS_TYPE_WIN2K:
    ip_header_out->ip_id=htons(morph_handle->ip_id++);
    ip_header_out->ip_off=0;
    ip_header_out->ip_ttl=128;
    tcp_header_out->th_win =0;
    tcp_header_out->th_ack = 0;
    if ((tcp_header_in->th_flags & TH_SYN) || (tcp_header_in->th_flags & TH_FIN)) {
      tcp_header_out->th_seq = htonl(new_sequence(morph_handle));
      tcp_header_out->th_ack = ntohl(tcp_header_in->th_seq);
      tcp_header_out->th_ack++;
      tcp_header_out->th_ack = htonl(tcp_header_out->th_ack);
      tcp_header_out->th_flags = TH_ACK | TH_RST;
    }
    break;
    
  case OS_TYPE_OPENBSD_33:
    ip_header_out->ip_id=rand_uint16(morph_handle->randomizer);
    ip_header_out->ip_ttl=255;
    ip_header_out->ip_off=htons(IP_DF);
    tcp_header_out->th_win = htons(0);
    if(tcp_header_in->th_flags & TH_SYN) {
      tcp_header_out->th_flags |= TH_ACK;
      tcp_header_out->th_ack = htonl(ntohl(tcp_header_in->th_seq) + 1); 
    } 

    if((tcp_header_in->th_flags & 0x29) == (TH_FIN | TH_PUSH | TH_URG)) {
      tcp_header_out->th_ack = htonl(ntohl(tcp_header_in->th_seq));
      tcp_header_out->th_flags |= TH_ACK;
    }
    
    break;
    
  case OS_TYPE_LINUX_24:
    ip_header_out->ip_id=0;
    ip_header_out->ip_ttl=64;
    ip_header_out->ip_off=htons(IP_DF);
    tcp_header_out->th_win = htons(5792);
    break;
  }

  ip_checksum(out_packet,ntohs(ip_header_out->ip_len));
  sentbytes = packetp_packet_inject(pp_ctx, out_packet);

  return(0);
}
